#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "util.h"
#include "doc.h"
#include "slda.h"
#include "evaluate.h"

double evaluate_pr(strct_doc* docs, int ndocs, double* yhat, const char* measure) {

	if (strcmp(measure, "mse") == 0) {

		double ret = 0;
		for(int i = 0; i < ndocs; i++)
			ret += SQR(docs[i].y - yhat[i]);
		return ret / ndocs;

	}
	else if (strcmp(measure, "pr2") == 0) {

		double bary = 0;
		for(int i = 0; i < ndocs; i++)
			bary += docs[i].y;
		bary /= ndocs;

		double s1 = 0, s2 = 0;
		for(int i = 0; i < ndocs; i++) {
			s1 += SQR(docs[i].y - yhat[i]);
			s2 += SQR(docs[i].y - bary);
		}

		return 1 - s1 / s2;

	}
	else {
		fprintf(stderr, "Ooops, pr error measure \"%s\" not supported.\n", measure);
		exit(1);
	}

}

double evaluate_perword_likelihood(strct_doc* docs, int ndocs, strct_slda* slda) {

	int K = slda->K;
	int V = slda->V;

	double ret = 0;
	int sum = 0;

	for(int i = 0; i < ndocs; i++) {
		sum += docs[i].word_count;
		for(int j = 0; j < docs[i].item_count; j++) {
			int v = docs[i].index[j];
			double p = 0;
			for(int k = 0; k < K; k++)
				p += docs[i].h->data[k] * slda->mu[k]->data[v];
			assert(p > 0);
			ret += docs[i].count[j] * log(p);
		}
	}

	return ret / sum;

}

// implementing the KM algorithm
// w: the weight matrix
// n: # of vertices on each side
// perm: max-weight matching, i --> perm[i], must be of size n
// return: total cost of the max-weight matching
double KM(double* w, int n, int* perm) {

	double *A, *B; // labels
	int *visited, *last;
	int *queue;
	int fr, re;
	int* _perm; // matching back

	assert(A = (double*)malloc(n * sizeof(double)));
	assert(B = (double*)malloc(n * sizeof(double)));
	assert(visited = (int*)malloc(2*n * sizeof(int)));
	assert(last = (int*)malloc(2*n * sizeof(int)));
	assert(queue = (int*)malloc(2*n * sizeof(int)));
	assert(_perm = (int*)malloc(n * sizeof(int)));

	// Initialization: A[i] = max w(i,j); B[i] = 0
	for(int i = 0; i < n; i++) {
		double maxw = -1e100;
		for(int j = 0; j < n; j++)
			if (w[ADDR2D(i, j, n)] > maxw) maxw = w[ADDR2D(i, j, n)];
		A[i] = maxw;
		B[i] = 0;
	}

	while (1) {

		// use BFS to find a perfect matching
		int unsuccess = -1;
		for(int i = 0; i < n; i++) {
			perm[i] = -1;
			_perm[i] = -1;
		}
		for(int root = 0; root < n; root++) {
			
			// Do BFS
			for(int i = 0; i < 2*n; i++) {
				visited[i] = 0;
				last[i] = -1;
			}
			fr = 0;
			re = 1;
			queue[fr] = root;
			visited[root] = 1;
			
			while (fr < re) {
				int v = queue[fr++];
				if (v < n) {
					// vertex on the left side
					int i = v;
					for(int j = 0; j < n; j++)
						if (!visited[n+j] && fabs(A[i] + B[j] - w[ADDR2D(i, j, n)]) < MINEPS) {
							visited[n+j] = 1;
							last[n+j] = i;
							queue[re++] = n+j;
							if (_perm[j] == -1) goto label0;
						}
				}
				else {
					// vertex on the right side
					int j = v - n;
					int i = _perm[j];
					if (!visited[i]) {
						visited[i] = 1;
						last[i] = n+j;
						queue[re++] = i;
					}
				}
			}
	
label0:

			if (fr < re) {
				// an augmentation path has been found: augment it, and continue the loop
				int v = queue[re-1];
				while (v != -1) {
					int lastv = last[v];
					perm[lastv] = v - n;
					_perm[v-n] = lastv;
					v = last[lastv];
				}
			}
			else {
				unsuccess = root;
				break;
			}

		}

		if (unsuccess == -1) {
			//successfully found a perfect matching
			break;
		}
		else {
			// alter the vertex labels
			double d = 1e100;
			for(int i = 0; i < n; i++)
				if (visited[i]) {
					for(int j = 0; j < n; j++)
						if (!visited[n+j]) {
							double value = A[i] + B[j] - w[ADDR2D(i, j, n)];
							if (value < d) d = value;
						}
				}
			for(int i = 0; i < n; i++) {
				if (visited[i]) A[i] -= d;
				if (visited[n+i]) B[i] += d;
			}
		}

	}

	free(A);
	free(B);
	free(visited);
	free(last);
	free(queue);
	free(_perm);

	double ret = 0;
	for(int i = 0; i < n; i++)
		ret += w[ADDR2D(i, perm[i], n)];
	return ret;

}

// compute \|x-y\|_p
double vecnorm(gsl_vector* x, gsl_vector* y, int n, double p) {

	double ret = 0;
	for(int i = 0; i < n; i++) {
		double delta = x->data[i] - y->data[i];
		if (fabs(delta) >= MINEPS) {
			ret += exp(p * log(fabs(delta)));
		}
	}
	
	if (fabs(ret) < MINEPS)
		return 0;
	else
		return exp(1.0/p * log(ret));

}

void evaluate_slda(strct_slda* slda, strct_slda* ref_slda, double p, int* perm, double* alphaacc, double* etaacc, double* muacc) {

	int K = slda->K;
	int V = slda->V;

	double *weight;
	assert(weight = (double*)malloc(SQR(K) * sizeof(double)));
	for(int i = 0; i < K; i++)
		for(int j = 0; j < K; j++)
			weight[ADDR2D(i, j, K)] = -vecnorm(slda->mu[i], ref_slda->mu[j], V, p);

	double sumweight = KM(weight, K, perm);

	if (muacc) *muacc = -sumweight / K;

	if (alphaacc) {
		*alphaacc = 0;
		for(int k = 0; k < K; k++) {
			double delta = slda->alpha->data[k] - ref_slda->alpha->data[perm[k]];
			if (fabs(delta) >= MINEPS) *alphaacc += exp(p * log(fabs(delta)));
		}
		if (fabs(*alphaacc) >= MINEPS)
			*alphaacc = exp(1.0/p * log(*alphaacc));
	}

	if (etaacc) {
		*etaacc = 0;
		for(int k = 0; k < K; k++) {
			double delta = slda->eta->data[k] - ref_slda->eta->data[perm[k]];
			if (fabs(delta) >= MINEPS) *etaacc += exp(p * log(fabs(delta)));
		}
		if (fabs(*etaacc) >= MINEPS)
			*etaacc = exp(1.0/p * log(*etaacc));
	}

	free(weight);

}
