#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>

#include "util.h"
#include "doc.h"
#include "slda.h"
#include "gibbsslda.h"

double compute_logprob(strct_doc* docs, int ndocs, strct_slda* slda, double* alpha, double* beta, double delta, double rho) {

	int K = slda->K;
	int V = slda->V;

	double ret = 0;

	// prior probability
	for(int k = 0; k < K; k++)
		for(int v = 0; v < V; v++) {
			assert(slda->mu[k]->data[v] > 0);
			ret += (beta[v]-1) * log(slda->mu[k]->data[v]);
		}
	for(int i = 0; i < ndocs; i++)
		for(int k = 0; k < K; k++) {
			assert(docs[i].h->data[k] > 0);
			ret += (alpha[k]-1) * log(docs[i].h->data[k]);
		}

	double t = 0;
	gsl_blas_ddot(slda->eta, slda->eta, &t);
	ret += -t / (2*SQR(rho));

	printf("prior ret = %lf, t = %lf, ", ret, t);

	double A = 0, B = 0, C = 0;

	gsl_vector *count = gsl_vector_alloc(K);
	for(int i = 0; i < ndocs; i++) {
		// p(y|eta, h)
		gsl_vector_set_all(count, 0);
		for(int j = 0; j < docs[i].word_count; j++) {
			assert(0 <= docs[i].z[j] && docs[i].z[j] < K);
			count->data[docs[i].z[j]] += 1.0/docs[i].word_count;
		}
		double y0 = 0;
		gsl_blas_ddot(slda->eta, count, &y0);
		ret += -SQR(y0-docs[i].y) / (2*SQR(delta));
		A += -SQR(y0-docs[i].y) / (2*SQR(delta));

		// p(z|h) and p(x|z)
		int p = 0;
		for(int j1 = 0; j1 < docs[i].item_count; j1++) {
			int v = docs[i].index[j1];
			for(int j2 = 0; j2 < docs[i].count[j1]; j2++) {
				int z = docs[i].z[p++];
				ret += log(docs[i].h->data[z]);
				B += log(docs[i].h->data[z]);
				ret += log(slda->mu[z]->data[v]);
				C += log(slda->mu[z]->data[v]);
			}
		}
	}

	printf("A = %lf, B = %lf, C = %lf, ", A, B, C);

	gsl_vector_free(count);

	return ret;

}

void update_z(strct_doc* docs, int ndocs, strct_slda* slda, double* alpha, double* beta, double delta, double rho, gsl_rng* rng, bool is_training) {

	puts("Updating z ...");

	int K = slda->K;
	int V = slda->V;

	gsl_vector* count = gsl_vector_alloc(K);
	gsl_vector* logprob = gsl_vector_alloc(K);
	double *prob;
	assert(prob = (double*)malloc(K * sizeof(double)));
	int percentage = 0;
	for(int i = 0; i < ndocs; i++) {
		if (i > ndocs * 0.01 * (percentage+1)) {
			putchar('.');
			fflush(stdout);
			percentage ++;
		}

		gsl_vector_set_all(count, 0);
		for(int j = 0; j < docs[i].word_count; j++)
			count->data[docs[i].z[j]] += 1.0/docs[i].word_count;
		double y1, y2;
		gsl_blas_ddot(slda->eta, count, &y1);

		// sampling
		int p = 0;
		for(int j1 = 0; j1 < docs[i].item_count; j1++) {
			int v = docs[i].index[j1];
			for(int j2 = 0; j2 < docs[i].count[j1]; j2++) {
				int lastz = docs[i].z[p];
				y1 -= 1.0/docs[i].word_count * slda->eta->data[lastz];
				for(int k = 0; k < K; k++) {
					y2 = y1 + 1.0/docs[i].word_count * slda->eta->data[k];
					logprob->data[k] = log(docs[i].h->data[k]) + log(slda->mu[k]->data[v]); 
					if (is_training) {
						logprob->data[k] += -1.0/(2*SQR(delta)) * SQR(docs[i].y-y2);
					}
				}
				double Z = gsl_vector_max(logprob);
				for(int k = 0; k < K; k++)
					prob[k] = exp(logprob->data[k] - Z);
				gsl_ran_discrete_t* g = gsl_ran_discrete_preproc(K, prob);
				docs[i].z[p] = gsl_ran_discrete(rng, g);
				gsl_ran_discrete_free(g);
				y1 += 1.0/docs[i].word_count * slda->eta->data[docs[i].z[p]];
				p ++;
			}
		}

	}

	puts("");

	gsl_vector_free(count);
	gsl_vector_free(logprob);
	free(prob);

}

void update_eta(strct_doc* docs, int ndocs, strct_slda* slda, double* alpha, double* beta, double delta, double rho, gsl_rng* rng) {

	puts("Updating eta ...");

	int K = slda->K;
	int V = slda->V;

	gsl_vector* mean = gsl_vector_alloc(K);
	gsl_matrix* lambda = gsl_matrix_alloc(K, K);

	gsl_vector_set_all(mean, 0);
	gsl_matrix_set_all(lambda, 0);
	for(int k = 0; k < K; k++)
		gsl_matrix_set(lambda, k, k, 1.0 / SQR(rho));

	gsl_vector* barz = gsl_vector_alloc(K);
	int percentage = 0;
	for(int i = 0; i < ndocs; i++) {
		if (i > ndocs * 0.01 * (percentage+1)) {
			putchar('.');
			fflush(stdout);
			percentage ++;
		}
		gsl_vector_set_all(barz, 0);
		for(int j = 0; j < docs[i].word_count; j++)
			barz->data[docs[i].z[j]] += 1.0 / docs[i].word_count;
		for(int k = 0; k < K; k++)
			mean->data[k] += docs[i].y * barz->data[k] / SQR(delta);
		for(int k1 = 0; k1 < K; k1++)
			for(int k2 = 0; k2 < K; k2++)
				*(gsl_matrix_ptr(lambda, k1, k2)) += barz->data[k1] * barz->data[k2] / SQR(delta);
	}
	puts("");

	// inverse the lambda matrix
	gsl_matrix* inv = gsl_matrix_alloc(K, K);
	gsl_permutation* perm = gsl_permutation_alloc(K);
	int s;
	gsl_linalg_LU_decomp(lambda, perm, &s);
	gsl_linalg_LU_invert(lambda, perm, inv);
	gsl_matrix_memcpy(lambda, inv);

	// update eta
	gsl_blas_dgemv(CblasNoTrans, 1.0, lambda, mean, 0, slda->eta);

	gsl_vector_free(mean);
	gsl_vector_free(barz);
	gsl_matrix_free(lambda);
	gsl_matrix_free(inv);
	gsl_permutation_free(perm);

}

void update_h(strct_doc* docs, int ndocs, strct_slda* slda, double* alpha, double* beta, double delta, double rho, gsl_rng* rng) {

	puts("Updating h ...");

	int K = slda->K;
	int V = slda->V;

	double* para;
	assert(para = (double*)malloc(K * sizeof(double)));
	for(int i = 0; i < ndocs; i++) {
		memcpy(para, alpha, K * sizeof(double));
		for(int j = 0; j < docs[i].word_count; j++) {
			assert(0 <= docs[i].z[j] && docs[i].z[j] < K);
			para[docs[i].z[j]] += 1;
		}
		gsl_ran_dirichlet(rng, K, para, docs[i].h->data); 
	}

	free(para);

}

void update_mu(strct_doc* docs, int ndocs, strct_slda* slda, double* alpha, double* beta, double delta, double rho, gsl_rng* rng) {

	puts("Updating mu ...");

	int K = slda->K;
	int V = slda->V;

	double* para;
	assert(para = (double*)malloc(K*V * sizeof(double)));

	for(int k = 0; k < K; k++)
		for(int v = 0; v < V; v++)
			para[ADDR2D(k, v, V)] = beta[v];

	for(int i = 0; i < ndocs; i++) {
		int p = 0;
		for(int j1 = 0; j1 < docs[i].item_count; j1++) {
			int v = docs[i].index[j1];
			for(int j2 = 0; j2 < docs[i].count[j1]; j2++) { 
				assert(0 <= docs[i].z[p] && docs[i].z[p] < K);
				para[ADDR2D(docs[i].z[p], v, V)] += 1;
				p ++;
			}
		}
	}

	for(int k = 0; k < K; k++)
		gsl_ran_dirichlet(rng, V, para + k*V, slda->mu[k]->data);

	free(para);

}

strct_slda* gibbsslda_fit(strct_doc* docs, int ndocs, int K, int V, double* alpha, double* beta, double delta, double rho, strct_slda* slda, double epsilon) {

	gsl_rng_env_setup();
	gsl_rng *rng = gsl_rng_alloc(gsl_rng_default);

	if (!alpha) {
		assert(alpha = (double*)malloc(K * sizeof(double)));
		for(int k = 0; k < K; k++)
			alpha[k] = DEFAULT_ALPHA; 
	}
	if (!beta) {
		assert(beta = (double*)malloc(V * sizeof(double)));
		for(int i = 0; i < V; i++)
			beta[i] = DEFAULT_BETA;
	}
	if (!slda) {
		slda = alloc_slda(K, V);
		init_slda(slda, rng); 
	}
	else {
		//override the alpha parameter
//		for(int k = 0; k < K; k++)
//			alpha[k] = slda->alpha->data[k];
		// smoothing mu 
		for(int k = 0; k < K; k++)
			for(int i = 0; i < V; i++)
				slda->mu[k]->data[i] = (slda->mu[k]->data[i] + 1e-3/V) / (1.0 + 1e-3);
	}

	printf("Initializing document latent variables ...\n");
	for(int i = 0; i < ndocs; i++) {
		init_hz(&(docs[i]), K, rng);
	}

	int iter = 0;
	double lastloss = -1e100;
	double logprob = -1e10;
	printf("Init loss: %lf\n", -compute_logprob(docs, ndocs, slda, alpha, beta, delta, rho));

	while (iter < MAX_ROUND && logprob > lastloss && fabs((logprob - lastloss) / lastloss) > epsilon) {
	
		update_z(docs, ndocs, slda, alpha, beta, delta, rho, rng, true);
		update_eta(docs, ndocs, slda, alpha, beta, delta, rho, rng);
		update_h(docs, ndocs, slda, alpha, beta, delta, rho, rng);
		update_mu(docs, ndocs, slda, alpha, beta, delta, rho, rng);

		// smooth the Mu matrix to avoid 0 probabilities
		for(int k = 0; k < K; k++)
			for(int i = 0; i < V; i++)
				slda->mu[k]->data[i] = (slda->mu[k]->data[i] + 1e-3/V) / (1.0 + 1e-3);

		// smooth the H matrix
		for(int i = 0; i < ndocs; i++)
			for(int k = 0; k < K; k++)
				docs[i].h->data[k] = (docs[i].h->data[k] + 1e-3/K) / (1.0 + 1e-3);

		lastloss = logprob;
		logprob = compute_logprob(docs, ndocs, slda, alpha, beta, delta, rho);
		printf("Iter %d: -log prob = %lf\n", ++iter, -logprob);

	}

	return slda;

}

double compute_truncated_logprob(strct_doc* docs, int ndocs, double delta, strct_slda* slda) {

	int K = slda->K;
	int V = slda->V;
	double ret = 0;

	gsl_vector* barz = gsl_vector_alloc(K);
	double A = 0, B = 0, C = 0;

	for(int i = 0; i < ndocs; i++) {
		//p(h|alpha) 
		for(int k = 0; k < K; k++) {
			ret += (slda->alpha->data[k]-1) * log(docs[i].h->data[k]);
			A += (slda->alpha->data[k]-1) * log(docs[i].h->data[k]);
		}
		//p(y|barz,eta)
/*		gsl_vector_set_all(barz, 0);
		for(int j = 0; j < docs[i].word_count; j++) 
			barz->data[docs[i].z[j]] += 1.0/docs[i].word_count;
		double y0 = 0;
		gsl_blas_ddot(barz, slda->eta, &y0);
		ret += -SQR(docs[i].y-y0) / (2*SQR(delta));
		B += -SQR(docs[i].y-y0) / (2*SQR(delta));*/
		//p(z|h) and p(x|z,mu)
		int p = 0;
		for(int j1 = 0; j1 < docs[i].item_count; j1++) {
			int v = docs[i].index[j1];
			for(int j2 = 0; j2 < docs[i].count[j1]; j2++) {
				int z = docs[i].z[p];
				ret += log(docs[i].h->data[z]);
				ret += log(slda->mu[z]->data[v]);
				B += log(docs[i].h->data[z]);
				C += log(slda->mu[z]->data[v]);
				p ++;
			}
		}
	}

	gsl_vector_free(barz);

	printf("A = %lf, B = %lf, C = %lf, ", A, B, C);

	return ret;

}

double* gibbsslda_predict(strct_doc* docs, int ndocs, double delta, strct_slda* slda, double epsilon) {

	int K = slda->K;
	int V = slda->V;

	gsl_rng_env_setup();
	gsl_rng* rng = gsl_rng_alloc(gsl_rng_default);

	printf("Initializing document latent variables ...\n");
	for(int i = 0; i < ndocs; i++) {
		init_hz(&(docs[i]), K, rng);
	}

	int iter = 0;
	double lastloss = -1e100;
	double logprob = -1e10;
	printf("Init loss: %lf\n", -compute_truncated_logprob(docs, ndocs, delta, slda));

	while (iter < MAX_ROUND && logprob > lastloss && fabs((logprob - lastloss) / lastloss) > epsilon) {
	
		update_z(docs, ndocs, slda, slda->alpha->data, NULL, delta, 0, rng, false);
		update_h(docs, ndocs, slda, slda->alpha->data, NULL, delta, 0, rng);

		// smooth h
		for(int i = 0; i < ndocs; i++)
			for(int k = 0; k < K; k++)
				docs[i].h->data[k] = (docs[i].h->data[k] + 1e-3/K) / (1.0 + 1e-3);

		lastloss = logprob;
		logprob = compute_truncated_logprob(docs, ndocs, delta, slda);
		printf("Iter %d: -logprob = %lf\n", ++iter, -logprob);

	}

	// make the predictions
	puts("Making the final predictions ...");
	double* yhat;
	assert(yhat = (double*)malloc(ndocs * sizeof(double)));
	gsl_vector* barz = gsl_vector_alloc(K);
	for(int i = 0; i < ndocs; i++) {
		gsl_vector_set_all(barz, 0);	
		for(int j = 0; j < docs[i].word_count; j++) {
			assert(0 <= docs[i].z[j] && docs[i].z[j] < K);
			barz->data[docs[i].z[j]] += 1.0/docs[i].word_count;
		}
		gsl_blas_ddot(barz, slda->eta, &(yhat[i]));
		docs[i].yhat = yhat[i];
	}
	gsl_vector_free(barz);

	return yhat;

}
